@@ -18,6 +18,7 @@ import datetime
 import eventlet
 import mock
 from oslo_config import cfg
+from six.moves import reload_module
 from stevedore import enabled
 import testtools
 
@@ -31,6 +32,7 @@ from blazar.notification import api as notifier_api
 from blazar.plugins import base
 from blazar.plugins import dummy_vm_plugin
 from blazar.plugins.oshosts import host_plugin
+from blazar import status
 from blazar import tests
 from blazar.utils.openstack import base as base_utils
 from blazar.utils import trusts
@@ -51,7 +53,7 @@ class FakePlugin(base.BasePlugin):
         return None
 
     def on_start(self, resource_id):
-        return 'Resorce %s should be started this moment.' % resource_id
+        return 'Resource %s should be started this moment.' % resource_id
 
     def on_end(self, resource_id):
         return 'Resource %s should be deleted this moment.' % resource_id
@@ -66,19 +68,28 @@ class FakePluginRaisesException(base.BasePlugin):
         raise Exception
 
     def on_start(self, resource_id):
-        return 'Resorce %s should be started this moment.' % resource_id
+        return 'Resource %s should be started this moment.' % resource_id
 
     def on_end(self, resource_id):
         return 'Resource %s should be deleted this moment.' % resource_id
 
 
+class FakeLeaseStatus(object):
+    @classmethod
+    def lease_status(cls, transition, result_in):
+        def decorator(func):
+            def wrapper(*args, **kwargs):
+                return func(*args, **kwargs)
+            return wrapper
+        return decorator
+
+
 class ServiceTestCase(tests.TestCase):
     def setUp(self):
         super(ServiceTestCase, self).setUp()
 
         self.cfg = cfg
         self.context = context
-        self.service = service
         self.enabled = enabled
         self.exceptions = exceptions
         self.eventlet = eventlet
@@ -88,6 +99,7 @@ class ServiceTestCase(tests.TestCase):
         self.trusts = trusts
         self.notifier_api = notifier_api
         self.base_utils = base_utils
+        self.status = status
 
         self.fake_plugin = self.patch(self.dummy_plugin, 'DummyVMPlugin')
 
@@ -103,6 +115,11 @@ class ServiceTestCase(tests.TestCase):
                                         'send_lease_notification')
 
         cfg.CONF.set_override('plugins', ['dummy.vm.plugin'], group='manager')
+
+        with mock.patch('blazar.status.lease.lease_status',
+                        FakeLeaseStatus.lease_status):
+            reload_module(service)
+        self.service = service
         self.manager = self.service.ManagerService()
 
         self.lease_id = '11-22-33'
@@ -209,46 +226,115 @@ class ServiceTestCase(tests.TestCase):
 
         self.assertFalse(event_update.called)
 
-    def test_event_all_okay(self):
+    def test_event_success(self):
         events = self.patch(self.db_api, 'event_get_first_sorted_by_filters')
         event_update = self.patch(self.db_api, 'event_update')
-        events.return_value = {'id': '111-222-333', 'time': self.good_date,
-                               'event_type': 'end_lease',
-                               'lease_id': self.lease_id}
+        events.return_value = {'id': '111-222-333',
+                               'time': self.good_date}
+        self.patch(eventlet, 'spawn_n')
 
         self.manager._event()
 
-        event_update.assert_called_once_with('111-222-333',
-                                             {'status': 'IN_PROGRESS'})
+        event_update.assert_called_once_with(
+            '111-222-333', {'status': status.event.IN_PROGRESS})
+
+    def test_event_spawn_fail(self):
+        events = self.patch(self.db_api, 'event_get_first_sorted_by_filters')
+        event_update = self.patch(self.db_api, 'event_update')
+        self.patch(eventlet, 'spawn_n').side_effect = Exception
+        events.return_value = {'id': '111-222-333',
+                               'time': self.good_date}
+
+        self.manager._event()
+
+        event_update.assert_has_calls([
+            mock.call('111-222-333', {'status': status.event.IN_PROGRESS}),
+            mock.call('111-222-333', {'status': status.event.ERROR})])
+
+    def test_exec_event_success(self):
+        event = {'id': '111-222-333',
+                 'event_type': 'start_lease',
+                 'lease_id': self.lease_id}
+        start_lease = self.patch(self.manager, 'start_lease')
+
+        self.manager._exec_event(event)
+
+        start_lease.assert_called_once_with(lease_id=event['lease_id'],
+                                            event_id=event['id'])
+        self.lease_get.assert_called_once_with(event['lease_id'])
         expected_context = self.trust_ctx.return_value
         self.fake_notifier.assert_called_once_with(
             expected_context.__enter__.return_value,
             notifier_api.format_lease_payload(self.lease),
-            'lease.event.end_lease')
+            'lease.event.start_lease')
 
-    def test_event_wrong_event_status(self):
-        events = self.patch(self.db_api, 'event_get_first_sorted_by_filters')
-        self.patch(self.db_api, 'event_update')
-        events.return_value = {'id': '111-222-333', 'time': self.good_date,
-                               'event_type': 'wrong_type',
-                               'lease_id': self.lease_id}
+    def test_exec_event_invalid_event_type(self):
+        event = {'id': '111-222-333',
+                 'event_type': 'invalid',
+                 'lease_id': self.lease_id}
 
         self.assertRaises(manager_ex.EventError,
-                          self.manager._event)
+                          self.manager._exec_event,
+                          event)
 
-    def test_event_wrong_eventlet_fail(self):
-        events = self.patch(self.db_api, 'event_get_first_sorted_by_filters')
+    def test_exec_event_retry(self):
+        event = {'id': '111-222-333',
+                 'event_type': 'start_lease',
+                 'lease_id': self.lease_id,
+                 'time': self.good_date}
+        start_lease = self.patch(self.manager, 'start_lease')
+        start_lease.side_effect = exceptions.InvalidStatus
         event_update = self.patch(self.db_api, 'event_update')
-        calls = [mock.call('111-222-333', {'status': 'IN_PROGRESS'}),
-                 mock.call('111-222-333', {'status': 'ERROR'})]
-        self.patch(eventlet, 'spawn_n').side_effect = Exception
-        events.return_value = {'id': '111-222-333', 'time': self.good_date,
-                               'event_type': 'end_lease',
-                               'lease_id': self.lease_id}
 
-        self.manager._event()
+        with mock.patch.object(datetime, 'datetime',
+                               mock.Mock(wraps=datetime.datetime)) as patched:
+            patched.utcnow.return_value = (self.good_date
+                                           + datetime.timedelta(seconds=1))
+            self.manager._exec_event(event)
 
-        event_update.assert_has_calls(calls)
+        start_lease.assert_called_once_with(lease_id=event['lease_id'],
+                                            event_id=event['id'])
+        event_update.assert_called_once_with(
+            event['id'], {'status': status.event.UNDONE})
+        self.lease_get.assert_not_called()
+
+    def test_exec_event_no_more_retry(self):
+        event = {'id': '111-222-333',
+                 'event_type': 'start_lease',
+                 'lease_id': self.lease_id,
+                 'time': self.good_date}
+        start_lease = self.patch(self.manager, 'start_lease')
+        start_lease.side_effect = exceptions.InvalidStatus
+        event_update = self.patch(self.db_api, 'event_update')
+
+        with mock.patch.object(datetime, 'datetime',
+                               mock.Mock(wraps=datetime.datetime)) as patched:
+            patched.utcnow.return_value = (self.good_date
+                                           + datetime.timedelta(days=1))
+            self.manager._exec_event(event)
+
+        start_lease.assert_called_once_with(lease_id=event['lease_id'],
+                                            event_id=event['id'])
+        event_update.assert_called_once_with(
+            event['id'], {'status': status.event.ERROR})
+        self.lease_get.assert_not_called()
+
+    def test_exec_event_handle_exception(self):
+        event = {'id': '111-222-333',
+                 'event_type': 'start_lease',
+                 'lease_id': self.lease_id,
+                 'time': self.good_date}
+        start_lease = self.patch(self.manager, 'start_lease')
+        start_lease.side_effect = Exception
+        event_update = self.patch(self.db_api, 'event_update')
+
+        self.manager._exec_event(event)
+
+        start_lease.assert_called_once_with(lease_id=event['lease_id'],
+                                            event_id=event['id'])
+        event_update.assert_called_once_with(
+            event['id'], {'status': status.event.ERROR})
+        self.lease_get.assert_not_called()
 
     def test_get_lease(self):
         lease = self.manager.get_lease(self.lease_id)
@@ -268,6 +354,7 @@ class ServiceTestCase(tests.TestCase):
         trust_id = 'exxee111qwwwwe'
         lease_values = {
             'id': self.lease_id,
+            'name': 'lease-name',
             'reservations': [{'id': '111',
                               'resource_id': '111',
                               'resource_type': 'virtual:instance',
@@ -291,6 +378,7 @@ class ServiceTestCase(tests.TestCase):
         trust_id = 'exxee111qwwwwe'
         lease_values = {
             'id': self.lease_id,
+            'name': 'lease-name',
             'reservations': [{'id': '111',
                               'resource_id': '111',
                               'resource_type': 'virtual:instance',
@@ -310,6 +398,7 @@ class ServiceTestCase(tests.TestCase):
     def test_create_lease_validate_created_events(self):
         lease_values = {
             'id': self.lease_id,
+            'name': 'lease-name',
             'reservations': [{'id': '111',
                               'resource_id': '111',
                               'resource_type': 'virtual:instance',
@@ -353,6 +442,7 @@ class ServiceTestCase(tests.TestCase):
     def test_create_lease_before_end_event_is_before_lease_start(self):
         lease_values = {
             'id': self.lease_id,
+            'name': 'lease-name',
             'reservations': [{'id': '111',
                               'resource_id': '111',
                               'resource_type': 'virtual:instance',
@@ -395,6 +485,7 @@ class ServiceTestCase(tests.TestCase):
 
     def test_create_lease_before_end_event_before_start_without_lease_id(self):
         lease_values = {
+            'name': 'lease-name',
             'reservations': [{'id': '111',
                               'resource_id': '111',
                               'resource_type': 'virtual:instance',
@@ -418,6 +509,7 @@ class ServiceTestCase(tests.TestCase):
         start_date = '2026-11-13 13:13'
         lease_values = {
             'id': self.lease_id,
+            'name': 'lease-name',
             'reservations': [{'id': '111',
                               'resource_id': '111',
                               'resource_type': 'virtual:instance',
@@ -435,6 +527,7 @@ class ServiceTestCase(tests.TestCase):
         before_end_date = '2026-11-15 13:13'
         lease_values = {
             'id': self.lease_id,
+            'name': 'lease-name',
             'reservations': [{'id': '111',
                               'resource_id': '111',
                               'resource_type': 'virtual:instance',
@@ -451,6 +544,7 @@ class ServiceTestCase(tests.TestCase):
     def test_create_lease_no_before_end_event(self):
         lease_values = {
             'id': self.lease_id,
+            'name': 'lease-name',
             'reservations': [{'id': '111',
                               'resource_id': '111',
                               'resource_type': 'virtual:instance',
@@ -489,6 +583,7 @@ class ServiceTestCase(tests.TestCase):
         before_end_date = '2026-11-14 10:13'
         lease_values = {
             'id': self.lease_id,
+            'name': 'lease-name',
             'reservations': [{'id': '111',
                               'resource_id': '111',
                               'resource_type': 'virtual:instance',
@@ -530,7 +625,8 @@ class ServiceTestCase(tests.TestCase):
         self.assertEqual('UNDONE', event['status'])
 
     def test_create_lease_wrong_date(self):
-        lease_values = {'start_date': '2025-13-35 13:13',
+        lease_values = {'name': 'lease-name',
+                        'start_date': '2025-13-35 13:13',
                         'end_date': '2025-12-31 13:13',
                         'trust_id': 'exxee111qwwwwe'}
 
@@ -539,16 +635,19 @@ class ServiceTestCase(tests.TestCase):
 
     def test_create_lease_wrong_format_before_end_date(self):
         before_end_date = '2026-14 10:13'
-        lease_values = {'start_date': '2026-11-13 13:13',
-                        'end_date': '2026-11-14 13:13',
-                        'before_end_date': before_end_date,
-                        'trust_id': 'exxee111qwwwwe'}
+        lease_values = {
+            'name': 'lease-name',
+            'start_date': '2026-11-13 13:13',
+            'end_date': '2026-11-14 13:13',
+            'before_end_date': before_end_date,
+            'trust_id': 'exxee111qwwwwe'}
 
         self.assertRaises(
             manager_ex.InvalidDate, self.manager.create_lease, lease_values)
 
     def test_create_lease_start_date_in_past(self):
         lease_values = {
+            'name': 'lease-name',
             'start_date':
             datetime.datetime.strftime(
                 datetime.datetime.utcnow() - datetime.timedelta(days=1),
@@ -562,6 +661,7 @@ class ServiceTestCase(tests.TestCase):
     def test_create_lease_unsupported_resource_type(self):
         lease_values = {
             'id': self.lease_id,
+            'name': 'lease-name',
             'reservations': [{'id': '111',
                               'resource_id': '111',
                               'resource_type': 'unsupported:type',
@@ -594,6 +694,32 @@ class ServiceTestCase(tests.TestCase):
         self.assertRaises(manager_ex.MissingTrustId,
                           self.manager.create_lease, lease_values)
 
+    def test_create_lease_without_required_params(self):
+        name_missing_values = {
+            'start_date': '2026-11-13 13:13',
+            'end_date': '2026-12-13 13:13',
+            'trust_id': 'trust1'}
+        start_missing_values = {
+            'name': 'name',
+            'end_date': '2026-12-13 13:13',
+            'trust_id': 'trust1'}
+        end_missing_values = {
+            'name': 'name',
+            'start_date': '2026-11-13 13:13',
+            'trust_id': 'trust1'}
+        resource_type_missing_value = {
+            'name': 'name',
+            'start_date': '2026-11-13 13:13',
+            'end_date': '2026-12-13 13:13',
+            'reservations': [{'min': 1, 'max': 3}],
+            'trust_id': 'trust1'
+            }
+
+        for value in [name_missing_values, start_missing_values,
+                      end_missing_values, resource_type_missing_value]:
+            self.assertRaises(manager_ex.MissingParameter,
+                              self.manager.create_lease, value)
+
     def test_update_lease_completed_lease_rename(self):
         lease_values = {'name': 'renamed'}
         target = datetime.datetime(2015, 1, 1)
@@ -601,7 +727,8 @@ class ServiceTestCase(tests.TestCase):
                                'datetime',
                                mock.Mock(wraps=datetime.datetime)) as patched:
             patched.utcnow.return_value = target
-            lease = self.manager.update_lease(self.lease_id, lease_values)
+            lease = self.manager.update_lease(lease_id=self.lease_id,
+                                              values=lease_values)
         self.lease_update.assert_called_once_with(self.lease_id, lease_values)
         self.assertEqual(lease, self.lease)
 
@@ -637,7 +764,8 @@ class ServiceTestCase(tests.TestCase):
                                'datetime',
                                mock.Mock(wraps=datetime.datetime)) as patched:
             patched.utcnow.return_value = target
-            self.manager.update_lease(self.lease_id, lease_values)
+            self.manager.update_lease(lease_id=self.lease_id,
+                                      values=lease_values)
         self.fake_plugin.update_reservation.assert_called_with(
             '593e7028-c0d1-4d76-8642-2ffd890b324c',
             {
@@ -691,7 +819,8 @@ class ServiceTestCase(tests.TestCase):
                                'datetime',
                                mock.Mock(wraps=datetime.datetime)) as patched:
             patched.utcnow.return_value = target
-            self.manager.update_lease(self.lease_id, lease_values)
+            self.manager.update_lease(lease_id=self.lease_id,
+                                      values=lease_values)
         self.fake_plugin.update_reservation.assert_called_with(
             '593e7028-c0d1-4d76-8642-2ffd890b324c',
             {
@@ -738,7 +867,7 @@ class ServiceTestCase(tests.TestCase):
             patched.utcnow.return_value = target
             self.assertRaises(
                 manager_ex.CantUpdateParameter, self.manager.update_lease,
-                self.lease_id, lease_values)
+                lease_id=self.lease_id, values=lease_values)
 
     def test_update_modify_reservations_without_reservation_id(self):
         lease_values = {
@@ -763,7 +892,7 @@ class ServiceTestCase(tests.TestCase):
             patched.utcnow.return_value = target
             self.assertRaises(
                 manager_ex.MissingParameter, self.manager.update_lease,
-                self.lease_id, lease_values)
+                lease_id=self.lease_id, values=lease_values)
 
     def test_update_lease_started_modify_end_date_without_before_end(self):
         def fake_event_get(sort_key, sort_dir, filters):
@@ -795,7 +924,8 @@ class ServiceTestCase(tests.TestCase):
                                'datetime',
                                mock.Mock(wraps=datetime.datetime)) as patched:
             patched.utcnow.return_value = target
-            self.manager.update_lease(self.lease_id, lease_values)
+            self.manager.update_lease(lease_id=self.lease_id,
+                                      values=lease_values)
         self.fake_plugin.update_reservation.assert_called_with(
             '593e7028-c0d1-4d76-8642-2ffd890b324c',
             {
@@ -844,7 +974,8 @@ class ServiceTestCase(tests.TestCase):
                                'datetime',
                                mock.Mock(wraps=datetime.datetime)) as patched:
             patched.utcnow.return_value = target
-            self.manager.update_lease(self.lease_id, lease_values)
+            self.manager.update_lease(lease_id=self.lease_id,
+                                      values=lease_values)
         self.fake_plugin.update_reservation.assert_called_with(
             '593e7028-c0d1-4d76-8642-2ffd890b324c',
             {
@@ -908,7 +1039,8 @@ class ServiceTestCase(tests.TestCase):
                                'datetime',
                                mock.Mock(wraps=datetime.datetime)) as patched:
             patched.utcnow.return_value = target
-            self.manager.update_lease(self.lease_id, lease_values)
+            self.manager.update_lease(lease_id=self.lease_id,
+                                      values=lease_values)
         self.fake_plugin.update_reservation.assert_called_with(
             '593e7028-c0d1-4d76-8642-2ffd890b324c',
             {
@@ -979,7 +1111,7 @@ class ServiceTestCase(tests.TestCase):
             patched.utcnow.return_value = target
             self.assertRaises(
                 exceptions.NotAuthorized, self.manager.update_lease,
-                self.lease_id, lease_values)
+                lease_id=self.lease_id, values=lease_values)
 
     def test_update_lease_started_modify_before_end_with_invalid_date(self):
         # before_end_date is greater than current end_date
@@ -1019,7 +1151,7 @@ class ServiceTestCase(tests.TestCase):
             patched.utcnow.return_value = target
             self.assertRaises(
                 exceptions.NotAuthorized, self.manager.update_lease,
-                self.lease_id, lease_values)
+                lease_id=self.lease_id, values=lease_values)
 
     def test_update_lease_started_modify_before_end_with_wrong_format(self):
         wrong_before_end_date = '12-21 14:00'
@@ -1058,11 +1190,12 @@ class ServiceTestCase(tests.TestCase):
             patched.utcnow.return_value = target
             self.assertRaises(
                 manager_ex.InvalidDate, self.manager.update_lease,
-                self.lease_id, lease_values)
+                lease_id=self.lease_id, values=lease_values)
 
     def test_update_lease_is_not_values(self):
-        lease_values = None
-        lease = self.manager.update_lease(self.lease_id, lease_values)
+        lease_values = {}
+        lease = self.manager.update_lease(lease_id=self.lease_id,
+                                          values=lease_values)
         self.lease_get.assert_called_once_with(self.lease_id)
         self.assertEqual(lease, self.lease)
 
@@ -1078,7 +1211,7 @@ class ServiceTestCase(tests.TestCase):
             patched.utcnow.return_value = target
             self.assertRaises(
                 exceptions.NotAuthorized, self.manager.update_lease,
-                self.lease_id, lease_values)
+                lease_id=self.lease_id, values=lease_values)
 
     def test_update_lease_not_started_start_date_before_current_time(self):
         lease_values = {
@@ -1092,7 +1225,7 @@ class ServiceTestCase(tests.TestCase):
             patched.utcnow.return_value = target
             self.assertRaises(
                 exceptions.NotAuthorized, self.manager.update_lease,
-                self.lease_id, lease_values)
+                lease_id=self.lease_id, values=lease_values)
 
     def test_update_lease_end_date_before_current_time(self):
         lease_values = {
@@ -1106,7 +1239,7 @@ class ServiceTestCase(tests.TestCase):
             patched.utcnow.return_value = target
             self.assertRaises(
                 exceptions.NotAuthorized, self.manager.update_lease,
-                self.lease_id, lease_values)
+                lease_id=self.lease_id, values=lease_values)
 
     def test_update_lease_completed_modify_dates(self):
         lease_values = {
@@ -1120,7 +1253,7 @@ class ServiceTestCase(tests.TestCase):
             patched.utcnow.return_value = target
             self.assertRaises(
                 exceptions.NotAuthorized, self.manager.update_lease,
-                self.lease_id, lease_values)
+                lease_id=self.lease_id, values=lease_values)
 
     def test_update_lease_start_date_event_not_found(self):
         events = self.patch(self.db_api, 'event_get_first_sorted_by_filters')
@@ -1131,8 +1264,7 @@ class ServiceTestCase(tests.TestCase):
         }
         self.assertRaises(exceptions.BlazarException,
                           self.manager.update_lease,
-                          self.lease_id,
-                          lease_values)
+                          lease_id=self.lease_id, values=lease_values)
 
     def test_update_lease_end_date_event_not_found(self):
         def fake_event_get(sort_key, sort_dir, filters):
@@ -1159,8 +1291,7 @@ class ServiceTestCase(tests.TestCase):
             patched.utcnow.return_value = target
             self.assertRaises(exceptions.BlazarException,
                               self.manager.update_lease,
-                              self.lease_id,
-                              lease_values)
+                              lease_id=self.lease_id, values=lease_values)
         event_update.assert_called_once_with(
             '2eeb784a-2d84-4a89-a201-9d42d61eecb1',
             {'time': datetime.datetime(2013, 12, 20, 13, 0)})
@@ -1224,6 +1355,30 @@ class ServiceTestCase(tests.TestCase):
         self.fake_plugin.on_end.assert_called_with('111')
         self.lease_destroy.assert_called_once_with(self.lease_id)
 
+    def test_delete_lease_after_starting_date_with_error_status(self):
+        def fake_event_get(sort_key, sort_dir, filters):
+            if filters['event_type'] == 'start_lease':
+                return {'id': 'fake', 'status': 'ERROR'}
+            elif filters['event_type'] == 'end_lease':
+                return {'id': 'fake', 'status': 'UNDONE'}
+            else:
+                return None
+        event_get = self.patch(db_api, 'event_get_first_sorted_by_filters')
+        event_get.side_effect = fake_event_get
+        fake_get_lease = self.patch(self.manager, 'get_lease')
+        fake_get_lease.return_value = self.lease
+        target = datetime.datetime(2013, 12, 20, 13, 30)
+        with mock.patch.object(datetime,
+                               'datetime',
+                               mock.Mock(wraps=datetime.datetime)) as patched:
+            patched.utcnow.return_value = target
+            self.manager.delete_lease(self.lease_id)
+
+        self.event_update.assert_called_once_with('fake',
+                                                  {'status': 'IN_PROGRESS'})
+        self.fake_plugin.on_end.assert_called_with('111')
+        self.lease_destroy.assert_called_once_with(self.lease_id)
+
     def test_start_lease(self):
         basic_action = self.patch(self.manager, '_basic_action')
 
@@ -1238,10 +1393,6 @@ class ServiceTestCase(tests.TestCase):
 
         self.manager.end_lease(self.lease_id, '1')
 
-        self.reservation_update.assert_called_with(
-            self.lease['reservations'][0]['id'],
-            {'status': 'completed'}
-        )
         self.trust_ctx.assert_called_once_with(self.lease['trust_id'])
         basic_action.assert_called_once_with(self.lease_id, '1', 'on_end',
                                              'deleted')
@@ -1260,6 +1411,8 @@ class ServiceTestCase(tests.TestCase):
 
     def test_basic_action_with_res_status(self):
         self.patch(self.manager, 'get_lease').return_value = self.lease
+        self.patch(self.status.reservation,
+                   'is_valid_transition').return_value = True
 
         self.manager._basic_action(self.lease_id, '1', 'on_end',
                                    reservation_status='IN_USE')
@@ -1276,6 +1429,8 @@ class ServiceTestCase(tests.TestCase):
             {'virtual:instance':
              {'on_start': self.fake_plugin.on_start,
               'on_end': raiseBlazarException}})
+        self.patch(self.status.reservation,
+                   'is_valid_transition').return_value = True
 
         self.patch(self.manager, 'get_lease').return_value = self.lease
 
